package org.cs.snc;

import javafx.embed.swing.SwingFXUtils;
import javafx.event.ActionEvent;
import javafx.fxml.FXML;
import javafx.fxml.FXMLLoader;
import javafx.fxml.Initializable;
import javafx.scene.control.*;
import javafx.scene.image.Image;
import javafx.scene.image.ImageView;
import javafx.scene.layout.AnchorPane;
import javafx.stage.FileChooser;
import javafx.stage.Stage;

import javax.imageio.*;
import java.io.File;
import java.io.IOException;
import java.net.URL;
import java.util.ArrayList;
import java.util.ResourceBundle;

public class FinalSceneController implements Initializable {

    // @FXML tag is required for any JavaFX GUI element that is
    // initialized at FXML load time instead of in the Java controller class
    @FXML private AnchorPane anchorPane;
    @FXML private Slider sliderImages;
    @FXML private ImageView imageView;
    @FXML private Label labelFinalFitness;
    @FXML private Label labelTotalTime;
    @FXML private Label labelNumOfGenerations;
    @FXML private Button btnStart;

    private ArrayList<Image> images;
    private long finalFitness;
    private int totalTimeInSeconds;
    private int numOfGenerations;


    /* ******************************************************************************************* */
    // Function Name: initialize                                                                   //
    // Return Type: void                                                                           //
    // Parameters: URL, ResourceBundle                                                             //
    // Description: The initialize function is required to override when implementing the JavaFX   //
    //              Initializable class. A Java class typically implements the Initializable class //
    //              if it is the controller of an fxml file. The initialize function is called     //
    //              once the contents of the fxml file that the Java class controls have been      //
    //              loaded. In this function, the min value of the image slider is set to 0 and    //
    //              the button to restart the algorithm is disabled.                               //
    /* ******************************************************************************************* */
    @Override
    public void initialize(URL url, ResourceBundle resourceBundle) {
        sliderImages.setMin(0);
        btnStart.setVisible(false);
    }

    /* ******************************************************************************************* */
    // Function Name: transferData                                                                 //
    // Return Type: void                                                                           //
    // Parameters: AfterAlgorithmData                                                              //
    // Description: This function is used to transfer data from RunScene to FinalScene. That data  //
    //              that is needed to display information after the algorithm is done running is   //
    //              stored in an AfterAlgorithmData object. The max value of the slider is set and //
    //              the information about the algorithm that was just run is displayed to the      //
    //              screen. The information is taken from the AfterAlgorithmData object.           //
    /* ******************************************************************************************* */
    public void transferData(AfterAlgorithmData data) {
        images = data.getImages();
        finalFitness = data.getFinalFitness();
        totalTimeInSeconds = data.getTotalTimeInSeconds();
        numOfGenerations = data.getNumOfGenerations();

        imageView.setFitWidth(images.get(0).getWidth());
        imageView.setFitHeight(images.get(0).getHeight());
        imageView.setImage(images.get(0));

        createSliderValues();
        createLabels();
    }

    /* ******************************************************************************************* */
    // Function Name: btnSaveClicked                                                               //
    // Return Type: void                                                                           //
    // Parameters: javafx.event.ActionEvent                                                        //
    // Description: This function is linked to a button via FinalSceneView.fxml. When clicked,     //
    //              a save file dialog is opened. The user can use this dialog to save the final   //
    //              generated image to their disk. The final image can only be saved as a PNG.     //
    /* ******************************************************************************************* */
    @FXML
    public void btnSaveClicked(ActionEvent actionEvent) {
        FileChooser fileChooser = new FileChooser();
        fileChooser.setTitle("Save");
        fileChooser.getExtensionFilters().addAll(new FileChooser.ExtensionFilter("PNG", "*.png"));

        Stage window = (Stage)((Button)actionEvent.getSource()).getScene().getWindow();
        File image_file = fileChooser.showSaveDialog(window);
        if(image_file != null) {
            try {
                ImageIO.write(SwingFXUtils.fromFXImage(images.get(images.size() - 1), null), "png", image_file);
                Alert alert = new Alert(Alert.AlertType.CONFIRMATION);
                alert.setContentText("Image save successfully.\nLocation: " + image_file.getAbsolutePath());
                alert.showAndWait();
            }
            catch(IOException ioe) {
                Alert alert = new Alert(Alert.AlertType.ERROR);
                alert.setContentText("Unable to save image. Try again.");
                alert.showAndWait();
            }
        }

        btnStart.setVisible(true);
    }

    /* ******************************************************************************************* */
    // Function Name: btnSaveClicked                                                               //
    // Return Type: void                                                                           //
    // Parameters: javafx.event.ActionEvent                                                        //
    // Description: This function is linked to a button via FinalSceneView.fxml. When clicked,     //
    //              ConfigSceneView.fxml is loaded to the screen and the user can start the        //
    //              again with a different picture.                                                //
    /* ******************************************************************************************* */
    @FXML
    public void btnStartClicked(ActionEvent actionEvent) throws IOException {
        Stage window = (Stage)((Button)actionEvent.getSource()).getScene().getWindow();

        AnchorPane pane = FXMLLoader.load(App.class.getResource("fxml/ConfigSceneView.fxml"));
        anchorPane.getChildren().setAll(pane);

        window.setWidth(1300);
        window.setHeight(800);
        window.setResizable(false);
        window.centerOnScreen();
    }


    private void createSliderValues() {
        sliderImages.setMax(images.size() - 1);
        sliderImages.valueProperty().addListener((observable, oldValue, newValue) -> imageView.setImage(images.get(newValue.intValue())));
    }


    private void createLabels() {
        labelFinalFitness.setText("" + finalFitness);
        labelNumOfGenerations.setText("" + numOfGenerations);

        if(totalTimeInSeconds < 3600) {
            int m = totalTimeInSeconds / 60;
            int s = totalTimeInSeconds - (m * 60);

            String minutes = String.format("%02d", m);
            String seconds = String.format("%02d", s);
            labelTotalTime.setText(minutes + ":" + seconds);
        }
        else {
            int h = totalTimeInSeconds / 3600;
            int m = (totalTimeInSeconds = (h * 60)) / 60;
            int s = totalTimeInSeconds - (m * 60) - (h * 3600);

            String hours = String.format("%02d", h);
            String minutes = String.format("%02d", m);
            String seconds = String.format("%02d", s);
            labelTotalTime.setText(hours + ":" + minutes + ":" + seconds);
        }
    }
}
