package org.cs.snc;

import java.awt.*;
import java.util.Random;

public class Individual {

    private final int[] polygons; // data is stored as [ x1, y1, x2, y2, ... , xn, yn, color] -> geneSize for triangles is 7
    private long fitness = -1;
    private static final Random random = new Random();


    /********************************* Constructors *********************************/
    /* ******************************************************************************************* */
    // This constructor is only called at the very beginning of the algorithm. It creates an       //
    // individual of random polygons.                                                              //
    //                                                                                             //
    // Steps for creating a random polygon:                                                        //
    //    1. The first (x,y) coordinate is created randomly                                        //
    //    2. Another point is generated but not completely randomly to prevent large polygons      //
    //          i.   A random side length l with 0 < l < 80 and radian value are generated         //
    //          ii.  The x value of the new point = (first x) + (side length * cos(radians))       //
    //          iii. The y value of the new point = (first y) + (side length * sin(radians))       //
    //    3. Step 2 is repeated until you have the specified number of vertices per polygon        //
    //    4. A random color is generated by generated random RGBA values.                          //
    //                                                                                             //
    // When the individual is done being created its fitness is calculated.                        //
    /* ******************************************************************************************* */
    public Individual() {
        int geneSize = (RunSceneController.numOfVertices * 2) + 1;
        this.polygons = new int[geneSize * RunSceneController.numOfPolygonsPerIndividual];

        for(int i = 0; i < (geneSize * RunSceneController.numOfPolygonsPerIndividual); i+=geneSize) {
            int originalX = random.nextInt(RunSceneController.imageWidth);
            int originalY = random.nextInt(RunSceneController.imageHeight);
            this.polygons[i] = originalX;
            this.polygons[i + 1] = originalY;

            double original_radians = 0.0;
            for(int j = 0; j < geneSize - 1; j+=2) {
                if(j == 2) {
                    int degrees = random.nextInt(360);
                    original_radians = (double)degrees * (Math.PI / 180);
                    int sideLength = random.nextInt(80);
                    this.polygons[i + j] = (int)(originalX + (sideLength * Math.cos(original_radians)));
                    this.polygons[i + j + 1] = (int)(originalY + (sideLength * Math.sin(original_radians)));
                }
                else {
                    int d_offset = random.nextInt(100) + 20;
                    double r_offset = (double)d_offset * (Math.PI / 180);
                    if(random.nextBoolean()) {
                        r_offset = 0 - r_offset;
                    }
                    double radians = original_radians + r_offset;
                    int sideLength = random.nextInt(80);
                    this.polygons[i + j] = (int)(originalX + (sideLength * Math.cos(radians)));
                    this.polygons[i + j + 1] = (int)(originalY + (sideLength * Math.sin(radians)));
                }
            }

            this.polygons[i + geneSize - 1] = (new Color(random.nextInt(255), random.nextInt(255), random.nextInt(255), random.nextInt(75) + 10)).getRGB();
        }
        calculateFitness();
    }


    // Copy constructor
    public Individual(Individual original) {
        this.polygons = new int[original.polygons.length];
        for(int i = 0; i < original.polygons.length; i++) {
            this.polygons[i] = original.polygons[i];
        }
        this.fitness = original.fitness;
    }


    // Creates a new Individual from genes taken from mother an father (only used for crossover)
    public Individual(Individual mother, Individual father) {
        this.polygons = new int[mother.polygons.length];
        int geneSize = (RunSceneController.numOfVertices * 2) + 1;
        if(RunSceneController.randomInheritance) {
            for(int i = 0; i < (geneSize * RunSceneController.numOfPolygonsPerIndividual); i+=geneSize) {
                if(Math.random() < 0.5) {
                    for(int j = 0; j < geneSize - 1; j+=2) {
                        this.polygons[i + j] = mother.polygons[i + j];
                        this.polygons[i + j + 1] = mother.polygons[i + j + 1];
                    }
                    this.polygons[i + geneSize - 1] = mother.polygons[i + geneSize - 1];
                }
                else {
                    for(int j = 0; j < geneSize - 1; j+=2) {
                        this.polygons[i + j] = father.polygons[i + j];
                        this.polygons[i + j + 1] = father.polygons[i + j + 1];
                    }
                    this.polygons[i + geneSize - 1] = father.polygons[i + geneSize - 1];
                }

                if(Math.random() < RunSceneController.mutationChance) {
                    mutatePolygon(i, geneSize);
                }
            }
        }
        else {
            int crossoverCutoff = (int)(RunSceneController.numOfPolygonsPerIndividual * RunSceneController.crossoverRatio);
            int count = 0;
            for(int i = 0; i < (geneSize * RunSceneController.numOfPolygonsPerIndividual); i+=geneSize) {
                if(count < crossoverCutoff) {
                    for(int j = 0; j < geneSize - 1; j+=2) {
                        this.polygons[i + j] = mother.polygons[i + j];
                        this.polygons[i + j + 1] = mother.polygons[i + j + 1];
                    }
                    this.polygons[i + geneSize - 1] = mother.polygons[i + geneSize - 1];
                }
                else {
                    for(int j = 0; j < geneSize - 1; j+=2) {
                        this.polygons[i + j] = father.polygons[i + j];
                        this.polygons[i + j + 1] = father.polygons[i + j + 1];
                    }
                    this.polygons[i + geneSize - 1] = father.polygons[i + geneSize - 1];
                }

                if(Math.random() < RunSceneController.mutationChance) {
                    mutatePolygon(i, geneSize);
                }
                count++;
            }
        }
        calculateFitness();
    }



    /******************************* Public Functions *******************************/
    public long getFitness() {
        if(this.fitness < 0) {
            calculateFitness();
        }
        return this.fitness;
    }


    public int[] getIndividual() {
        return this.polygons;
    }



    /****************************** Private Functions *******************************/
    private void calculateFitness() {
        this.fitness = FitnessCalc.getFitness(this.polygons);
    }


    private void mutatePolygon(int index, int geneSize) {
        for(int j = 0; j < geneSize - 1; j+=2) {
            this.polygons[index + j] += (random.nextInt(20) - 10);
            this.polygons[index + j + 1] += (random.nextInt(20) - 10);
        }

        Color color = new Color(this.polygons[index + geneSize - 1], true);
        int red = getValidColor(color.getRed() + (random.nextInt(30) - 15));
        int green = getValidColor(color.getGreen() + (random.nextInt(30) - 15));
        int blue = getValidColor(color.getBlue() + (random.nextInt(30) - 15));
        int alpha = getValidAlpha(color.getAlpha() + (random.nextInt(30) - 15));

        this.polygons[index + geneSize - 1] = (new Color(red, green, blue, alpha)).getRGB();
    }


    private int getValidColor(int c) {
        if(c < 0) {
            return 0;
        }
        else {
            return Math.min(c, 255);
        }
    }


    private int getValidAlpha(int a) {
        if(a < 10) {
            return 10;
        }
        else {
            return Math.min(a, 85);
        }
    }
}
