package org.cs.snc;

import javafx.application.Platform;
import javafx.embed.swing.SwingFXUtils;
import javafx.event.ActionEvent;
import javafx.fxml.FXML;
import javafx.fxml.FXMLLoader;
import javafx.fxml.Initializable;
import javafx.scene.control.Button;
import javafx.scene.control.Label;
import javafx.scene.image.Image;
import javafx.scene.image.ImageView;
import javafx.scene.layout.AnchorPane;
import javafx.stage.Stage;

import java.awt.Graphics2D;
import java.awt.RenderingHints;
import java.awt.image.BufferedImage;
import java.io.IOException;
import java.net.URL;
import java.util.ArrayList;
import java.util.ResourceBundle;
import java.util.Timer;
import java.util.TimerTask;
import java.util.concurrent.Callable;
import java.util.concurrent.FutureTask;

public class RunSceneController implements Initializable {

    // @FXML tag is required for any JavaFX GUI element that is
    // initialized at FXML load time instead of in the Java controller class
    @FXML private AnchorPane anchorPane;
    @FXML private ImageView imgViewImported;
    @FXML private ImageView imgViewGenerated;
    @FXML private Button btnStartStop;
    @FXML private Label labelElapsedTime;
    @FXML private Label labelNumOfGenerations;
    @FXML private Label labelCurrentFitness;
    @FXML private Label labelStartingFitness;

    private Timer timer;
    private int timeInSeconds = 0;
    private String timeLabelString = "";
    private int generationCount = 0;
    private long currentFitness;
    private long startingFitness;

    private boolean algorithmIsRunning = false;
    private ArrayList<Image> evolution;

    private static final int drawImageWidth = 400;
    private static int drawImageHeight = 0;

    public static int imageWidth = 0;
    public static int imageHeight = 0;

    public static int numOfIndividualsPerGeneration;
    public static int numOfPolygonsPerIndividual;
    public static int numOfVertices;

    public static double mutationChance;
    public static double crossoverRatio;
    public static boolean randomInheritance;
    public static boolean elitism;

    public static BufferedImage solution;


    /* ******************************************************************************************* */
    // Function Name: initialize                                                                   //
    // Return Type: void                                                                           //
    // Parameters: URL, ResourceBundle                                                             //
    // Description: The initialize function is required to override when implementing the JavaFX   //
    //              Initializable class. A Java class typically implements the Initializable class //
    //              if it is the controller of an fxml file. The initialize function is called     //
    //              once the contents of the fxml file that the Java class controls have been      //
    //              loaded. In this function, the style of the start button is set.                //
    /* ******************************************************************************************* */
    @Override
    public void initialize(URL url, ResourceBundle resourceBundle) {
        btnStartStop.setStyle("-fx-background-color: green; -fx-font-size: 20; -fx-background-radius: 15; -fx-text-fill: white;");
    }

    /* ******************************************************************************************* */
    // Function Name: transferData                                                                 //
    // Return Type: void                                                                           //
    // Parameters: AlgorithmData                                                                   //
    // Description: This function is used to transfer data from ConfigScene to RunScene. The data  //
    //              that is needed to run the algorithm is stored in an AlgorithmData object. That //
    //              object is passed to this function and the contents are used to set the values  //
    //              used in this scene.                                                            //
    /* ******************************************************************************************* */
    public void transferData(AlgorithmData data) {
        numOfIndividualsPerGeneration = data.getNumOfIndividualsPerGeneration();
        numOfPolygonsPerIndividual = data.getNumOfPolygonsPerIndividual();
        numOfVertices = data.getNumOfVertices();

        mutationChance = data.getMutationChance();
        crossoverRatio = data.getCrossoverRatio();
        randomInheritance = data.getRandomInheritance();
        elitism = data.getElitism();

        solution = data.getImportedBuffIm();
        imgViewImported.setImage(SwingFXUtils.toFXImage(data.getImportedBuffIm(), null));
    }

    /* ******************************************************************************************* */
    // Function Name: btnStartStopClicked                                                          //
    // Return Type: void                                                                           //
    // Parameters: javafx.event.ActionEvent                                                        //
    // Description: This function is linked to a button via RunSceneView.fxml. That button serves  //
    //              two purposes. If the algorithm is not running, the button says start and is    //
    //              green. If the algorithm is running, then the button says stop and is red. This //
    //              prevents the user from running two algorithms at the same time.                //
    /* ******************************************************************************************* */
    @FXML
    public void btnStartStopClicked(ActionEvent actionEvent) throws IOException {
        if(!algorithmIsRunning) {
            algorithmIsRunning = true;
            btnStartStop.setText("Stop");
            btnStartStop.setStyle("-fx-background-color: red; -fx-font-size: 20; -fx-background-radius: 15; -fx-text-fill: white;");
            runAlgorithm();
        }
        else {
            timer.cancel();
            algorithmIsRunning = false;

            Stage window = (Stage)((Button)actionEvent.getSource()).getScene().getWindow();
            window.setWidth(1000);
            window.setHeight(600);
            window.setResizable(false);
            window.centerOnScreen();

            FXMLLoader fxmlLoader = new FXMLLoader(App.class.getResource("fxml/FinalSceneView.fxml"));
            AnchorPane pane = fxmlLoader.load();

            FinalSceneController finalSceneController = fxmlLoader.getController();
            AfterAlgorithmData data = new AfterAlgorithmData(evolution, currentFitness, timeInSeconds, generationCount);
            finalSceneController.transferData(data);

            anchorPane.getChildren().setAll(pane);
        }
    }


    /* ******************************************************************************************* */
    // Function Name: runAlgorithm                                                                 //
    // Return Type: void                                                                           //
    // Parameters: None                                                                            //
    // Description: This is where the main algorithm runs. The algorithm does not run on the main  //
    //              application thread. Instead, a new thread is created so the GUI still responds //
    //              to user actions. A java.util.Timer is scheduled using a java.util.TimerTask    //
    //              to keep track of how much time has elapsed during algorithm run time. Because  //
    //              the algorithm runs a separate thread from the main application, it cannot      //
    //              GUI elements itself. Instead, it must call Platform.runLater(Runnable), which  //
    //              requests that the given Runnable be run on the main application thread. A      //
    //              Runnable called GUIUpdater is given to Platform.runLater that updates the      //
    //              information about the currently running algorithm. The evolution array keeps   //
    //              every twentieth generated image so the user can see the evolution in the next  //
    //              scene.
    /* ******************************************************************************************* */
    private void runAlgorithm() {
        evolution = new ArrayList<>();

        TimerTask timerTask = new TimerTask() {
            @Override
            public void run() {
                timeInSeconds++;
                if (timeInSeconds < 3600) {
                    int m = timeInSeconds / 60;
                    int s = timeInSeconds - (m * 60);

                    String minutes = String.format("%02d", m);
                    String seconds = String.format("%02d", s);
                    timeLabelString = minutes + ":" + seconds;
                } else {
                    int h = timeInSeconds / 3600;
                    int m = (timeInSeconds - (h * 60)) / 60;
                    int s = timeInSeconds - (m * 60) - (h * 3600);

                    String hours = String.format("%02d", h);
                    String minutes = String.format("%02d", m);
                    String seconds = String.format("%02d", s);
                    timeLabelString = hours + ":" + minutes + ":" + seconds;
                }

                Runnable GUIUpdater = () -> labelElapsedTime.setText(timeLabelString);
                Platform.runLater(GUIUpdater);
            }
        };
        timer = new Timer();
        timer.schedule(timerTask, 0, 1000);

        Thread runThread = new Thread(() -> {
            Individual[] current = new Individual[numOfIndividualsPerGeneration];
            Runnable GUIUpdater = () -> {
                labelNumOfGenerations.setText("" + generationCount);
                labelCurrentFitness.setText("" + currentFitness);

                BufferedImage generatedImage = FitnessCalc.draw(current[0].getIndividual());
                BufferedImage resized = new BufferedImage(drawImageWidth, drawImageHeight, BufferedImage.TYPE_3BYTE_BGR);
                Graphics2D g = resized.createGraphics();
                g.setRenderingHint(RenderingHints.KEY_INTERPOLATION, RenderingHints.VALUE_INTERPOLATION_BILINEAR);
                g.drawImage(generatedImage, 0, 0, drawImageWidth, drawImageHeight, null);
                g.dispose();

                imgViewGenerated.setImage(SwingFXUtils.toFXImage(resized, null));

                if(generationCount % 20 == 0) {
                    evolution.add(SwingFXUtils.toFXImage(resized, null));
                }
            };

            BufferedImage scaledSolution = scaleImage();
            FitnessCalc.setSolution(scaledSolution);

            for(int i = 0; i < current.length; i++) {
                current[i] = new Individual();
            }
            sort(current);
            currentFitness = startingFitness = current[0].getFitness();
            Platform.runLater(GUIUpdater);
            Platform.runLater(() -> labelStartingFitness.setText("" + startingFitness));

            int elitismOffset = elitism ? 2 : 0;
            int numOfTasks = numOfIndividualsPerGeneration - elitismOffset;

            while(algorithmIsRunning) {
                ArrayList<FutureTask<Individual>> tasks = new ArrayList<>();
                for(int i = 0; i < numOfTasks; i++) {
                    Callable<Individual> callable = new EvolveThread(current);
                    tasks.add(new FutureTask<>(callable));
                }
                for(FutureTask<Individual> f: tasks) {
                    Thread thread = new Thread(f);
                    thread.start();
                }

                Individual[] next = new Individual[numOfIndividualsPerGeneration];
                if(elitism) {
                    next[0] = new Individual(current[0]);
                    next[1] = new Individual(current[1]);
                }

                int index = elitismOffset;
                int count = 0;
                while(index < numOfIndividualsPerGeneration) {
                    try {
                        next[index] = tasks.get(count).get();
                        index++;
                        count++;
                    }
                    catch(Exception e) {
                        e.printStackTrace();
                    }
                }

                for(int i = 0; i < numOfTasks; i++) {
                    tasks.clear();
                    tasks.trimToSize();
                }
                System.gc();

                sort(next);
                for(int i = 0; i < numOfIndividualsPerGeneration; i++) {
                    current[i] = next[i];
                }
                generationCount++;
                currentFitness = current[0].getFitness();
                Platform.runLater(GUIUpdater);
            }
        });
        runThread.start();
    }



    /******************************* Private Functions ******************************/
    private void sort(Individual[] array) {
        quickSort(array, 0, array.length - 1);
    }


    /* ******************************************************************************************* */
    // Function Name: scaleImage                                                                   //
    // Return Type: java.awt.image.BufferedImage                                                   //
    // Parameters: None                                                                            //
    // Description: This function scales the imported image down to a width of 250 and a height    //
    //              proportional to the original height. This gives the algorithm significantly    //
    //              less pixels to compare while still having an accurate representation of the    //
    //              original image.                                                                //
    /* ******************************************************************************************* */
    private BufferedImage scaleImage() {
        int originalImageWidth = solution.getWidth();
        int originalImageHeight = solution.getHeight();

        double scale = (double) originalImageWidth / (double) originalImageHeight;
        imageWidth = 250;
        imageHeight = (int)(imageWidth / scale);

        drawImageHeight = (int)(drawImageWidth / scale);

        BufferedImage scaledImage = new BufferedImage(imageWidth, imageHeight, BufferedImage.TYPE_3BYTE_BGR);
        Graphics2D g = scaledImage.createGraphics();
        g.drawImage(solution, 0, 0, imageWidth, imageHeight, null);
        g.dispose();

        return scaledImage;
    }



    /****************************** QuickSort Functions *****************************/
    private void quickSort(Individual[] array, int low, int high) {
        if(low < high) {
            int index = partition(array, low, high);
            quickSort(array, low, index - 1);
            quickSort(array, index + 1, high);
        }
    }


    private int partition(Individual[] array, int low, int high) {
        long pivot = array[high].getFitness();
        int i = low - 1;
        for(int j = low; j <= high; j++) {
            if(array[j].getFitness() < pivot) {
                i++;
                swap(array, i, j);
            }
        }
        swap(array, i + 1, high);
        return (i + 1);
    }


    private void swap(Individual[] array, int i, int j) {
        Individual temp = array[i];
        array[i] = array[j];
        array[j] = temp;
    }
}
